#ifndef HX_3D_WAND_BASE_H
#define HX_3D_WAND_BASE_H

#include <Inventor/SbLinear.h>
#include <mclib/McHandle.h>
#include <Inventor/nodes/SoSwitch.h>
#include <hxcore/api.h>


#pragma warning(push)

// 'identifier' : class 'type' needs to have dll-interface to be used by clients of class 'type2'
#pragma warning(disable:4251)

/** Macro to access the wand from a SoTrackerEvent or a SoControllerButtonEvent
*/
#define GET_WAND(e) \
        dynamic_cast<Hx3DWandBase*>((Hx3DWandBase*)(e)->trackerInfo().getData())

/** Base class representing a 3D wand in XScreen. If you are writing
 *  an event callback which reacts on VR events in the application, i.e., on
 *  SoTrackerEvent or SoControllerButtonEvent, you can access the wand
 *  thanks to the GET_WAND(e) macro:
 *  @code
 *  SoTrackerEvent* e = (SoTrackerEvent*) action->getEvent();
 *  Hx3DWandBase* wand = GET_WAND(e);
 *
 *  or
 *
 *  SoControllerButtonEvent* e = (SoControllerButtonEvent*) action->getEvent();
 *  Hx3DWandBase* wand = GET_WAND(e);
 *  @endcode
 *  You can then ask the wand for its current position or
 *  orientation. You can also highlight it in order to give some
 *  visual feedback.
 *
 *  In addition, the class provides the two methods sceneXform()
 *  and sceneXformInverse(). You can use these methods to convert between
 *  the physical coordinates used in an XScreen configuration file and the scene
 *  coordinates, e.g., the coordinates a data set is defined in.
 */
class HXCORE_API Hx3DWand
{
  public:

    /** @brief Sets the inventor geometry used to display the wand
     *  in normal state, as well as the top of the wand. 
     * 
     *  The origin is assumed to be at (0,0,0). This can also be specified in a
     *  permanent way in the XScreen configuration file.
     *  @param wand    The node representing wand geometry (if NULL a default wand is created)
     *  @param hotSpot The hotspot of the wand 
     */
    void setGeometry(SoNode* wand, const SbVec3f& hotSpot);

    /** @brief Sets the inventor geometry used to display the wand in highlighted state
     *
     *  This can also be specified in a permanent way in the XScreen configuration file. 
     *  @param wand   The node representing wand geometry (if NULL a default wand is created)
     */
    void setHighlightGeometry(SoNode* wand);

    /** @brief Sets the scale factor applied to the wand geometry. 
     *
     *  The wand geometry will be scaled by this factor. The factor allows
     *  you to adjust the size of the wand so that it matches the
     *  VR coordinates used in the current XScreen configuration file. For
     *  example, if your screen width is 200 (cm) and you are
     *  using a wand geometry with the hot spot at (0,0,-1), a
     *  wand scale factor of 100 makes the wand half as long as your screen
     *  @param scaleFactor The scale factor
     */
    void setWandScale(float scaleFactor);

    /** @brief Returns the scale factor of the wand.
     *
     * The scale factor can be defined in the XScreen configuration file.
     * If it is not specified there, a default factor will be used which 0.16 times the
     * length of the first screen in defined in the config file. 
     * @return The scale factor of the wand.
     */
    float getWandScale() const { return m_wandScale; }

    /** @brief Highlights the wand.
     * 
     *  Can be used in custom nodes to give
     *  some visual feedback. For example, if the top of the wand
     *  is inside some box this method could be called so that the
     *  user knows that he can move the box. 
     *  @param enable  true for enable, false for disable
     */
    virtual void setHighlight(bool enable);

    /** @brief Returns the current state of a button of the VR input device.
     *  
     * @param butIdx The button index
     * @return true for pressed, false for released 
     */ 
    bool getButton(int butIdx) const;

    /** @brief Returns the current combined state of all buttons of the VR input device.
     *
     *  @return buttons status 
     */
    int getButtonStatus() const { return m_buttonStatus; }

    /** @brief Returns true if the specified button was pressed since the
     *
     * scene graph was traversed the last time by the handle event action. 
     * @param butIdx The button index
     * @return true for pressed
     */
    bool wasButtonPressed(int butIdx) const;

    /** @brief Returns true if the specified button was released since the
     *  scene graph was traversed the last time by the handle event action. 
     * @param butIdx The button index
     * @return true for released
     */
    bool wasButtonReleased(int butIdx) const;

    /** @brief Returns the origin of the 3D wand.
     *
     * If @c scene is true the origin is reported in scene coordinates, i.e., the value
     * depends on the current viewing transformation. Otherwise the
     * origin is reported in VR coordinates, i.e., in the same coordinate system used
     * in the XScreen configuration file. 
     * @param scene true for scene coordinates, false for VR coordinates
     * @return The origin 
     */
    const SbVec3f& origin(bool scene=true) const { return scene ? m_sceneOrigin : m_vrOrigin; }

    /** @brief Returns the top of the 3D wand. 
     *
     * If @c scene is true the top is reported in scene coordinates, i.e., the value
     * depends on the current viewing transformation. Otherwise the
     * top is reported in VR coordinates, i.e., in the same
     * coordinate system used in the XScreen configuration file. 
     * @param scene true for scene coordinates, false for VR coordinates
     * @return The top
     */
    const SbVec3f& top(bool scene=true) const { return scene ? m_sceneTop : m_vrTop; }

    /** @brief Returns the current orientation of the wand. 
     *
     *  If @c scene is true the orientation is reported relative to the scene
     *  coordinate system. Otherwise it is reported relative to the VR
     *  coordinate system, i.e., the coordinate system used in the
     *  XScreen configuration file. This method can be used to pick and move
     *  an object with the wand. In order to do so you need to remember
     *  the current wand top position and the current orientation at
     *  the time the interaction starts, i.e., when a wand button is
     *  pressed:
     *
     *  @code
     *  refRotInverse = wand->orientation().inverse();
     *  refPos = wand->origin();
     *  @endcode
     *
     *  Then, when the wand position and orientation changes you can
     *  compute a transformation matrix as follows:
     *
     *  @code
     *  SbMatrix T1; T1.setTranslate(-refPos);
     *  SbMatrix R; R.setRotate(refRotInverse*wand->orientation());
     *  SbMatrix T2; T2.setTranslate(wand->origin());
     *  SbMatrix deltaXform = T1*R*T2;
     *  @endcode
     *
     *  For a concrete example see the module MyVRDemo in the XScreen
     *  demo package. 
     *  @param scene true for scene orientation, false for VR orientation
     *  @returns The current orientation of the wand.   
     */
    const SbRotation& orientation(bool scene=true) const { return scene ? m_sceneOrientation : m_vrOrientation; }

    /** @brief Returns the length of the wand. If @c scene is true the length is
     *  reported in scene coordinates. Otherwise it is reported in VR
     *  coordinates. The length is the distance between the wand origin
     *  and the top of the wand.
     */
    float length(bool scene=true) const { return scene ? m_sceneLength : m_vrLength; }

    /** @brief Returns the matrix which transforms between scene coordinates
     *  and VR coordinates, i.e., the coordinates used in the XScreen
     *  configuration file. For example, in order to get the top position of
     *  the wand in this physical coordinate system use
     *
     *  \code
     *  SbVec3f topInDeviceCoords;
     *  wand->sceneXform().multVecMatrix(wand->top(),topInDeviceCoords);
     *  \endcode 
     * @returns The transform matrix.
     */
    const SbMatrix& sceneXform() const { return *m_sceneToVRTransform; }

    /** @brief Returns the inverse of sceneXform().
     * 
     * This matrix transforms from device coordinates as defined in the XScreen configuration file
     * to scene coordinates.
     * @returns The inverse of sceneXform().
     */
    const SbMatrix& sceneXformInverse() const { return *m_vrToSceneTransform; }

    /** @brief Updates the 3D wand position.
     * 
     * @param pos The new position
     * @param rot The new rotation
     * @param butStat The new combined button status
     */
    virtual void update(const SbVec3f& pos, const SbRotation& rot, int butStat);

    /** @brief Returns the HotSpot
     * 
     * @return The HotSpot
     */
    const SbVec3f& getHotSpot() const { return m_hotSpot; }

    /** @brief Creates a default Wand geometry node
     * 
     * @return The node representing a wand
     */
    static SoSeparator* getDefaultWandGeometry();
    
protected:
    /** Constructor 
     */
    Hx3DWandBase();

    /** Destructor 
     */
    virtual ~Hx3DWandBase();

    /** @brief Creates a default Wand based on a default geometry node
     * 
     * @param highlight true for getting highlighted geometry
     * @return The node
     */
    static SoSeparator* createDefaultWand( bool highlight);

    /**
     * The hot spot of the wand   
     */
    SbVec3f m_hotSpot;

    /**
     * The scale factor of the wand   
     */
    float m_wandScale;

    /**
     * The node for switching between highlight ON/OFF
     */
    McHandle<SoSwitch> m_highlightSwitch;

    /**
     * Code for the two (highlighted and not) geometries
     */
    enum WandGeom {
        DEFAULT_GEOM = 0,    ///< Default geometry
        HIGHTLIGHT_GEOM = 1  ///< Highlight geometry
    };

    /**
     * The length in scene coordinates
     */
    float m_sceneLength;

    /**
     * The top in scene coordinates
     */
    SbVec3f m_sceneTop;

    /**
     * The origin in scene coordinates
     */
    SbVec3f m_sceneOrigin;

    /**
     * The orientation in scene coordinates
     */
    SbRotation m_sceneOrientation;

    /**
     * The length in VR coordinates
     */
    float m_vrLength;

    /**
     * The top in VR coordinates
     */
    SbVec3f m_vrTop;

    /**
     * The origin in VR coordinates
     */
    SbVec3f m_vrOrigin;

    /**
     * The orientation in VR coordinates
     */
    SbRotation m_vrOrientation;

    /**
     * The scene to VR transform
     */
    const SbMatrix* m_sceneToVRTransform;

    /**
     * The inverse of scene to VR transform
     */
    const SbMatrix* m_vrToSceneTransform;

    /**
     * The buttons (all) combined status
     */
    int m_buttonStatus;

    /**
     * The buttons (all) combined status
     * sampled during last scene graph traversal
     */
    int m_lastButtonStatus;
};

#pragma warning(pop)

#endif
