/*
Copyright 1999-2016 FEI, SAS
Copyright 1995-2016, ZIB Berlin
ALL RIGHTS RESERVED.

Avizo is a registered trademark of FEI, SAS
Amira is a registered trademark of ZIB.

Avizo and Amira are being jointly developed by Zuse Institute Berlin and FEI, SAS.
HardCopy, MeshViz, VolumeViz, TerrainViz are marks of FEI, SAS
FEI, SAS is a source licensee of OpenGL, Open Inventor
from Silicon Graphics, Inc. OpenGL and Open Inventor are
registered trademark of Silicon Graphics, Inc. All other products and
company names are trademarks or registered trademarks of
their respective companies.

FEI, SAS accepts no liability to any
party for loss or damage caused by errors or omissions or by
statements of any kind in this publication, whether such errors
result from accident, negligence or any other cause. FEI assumes
no liability for incidental or consequential damages arising
from the use of information in this publication. FEI provides no
warranties regarding the information contained in this
publication, whether expressed, implied, or statutory, including
implied warranties of merchantability or fitness for a
particular purpose.

FEI also does not assume any liability for links from this
publication to publications from other companies, organizations
or individuals. FEI does not imply endorsement of any specific
company, organization or individual through offering links to
this content. All information provided within this publication
is subject to change at any time and is provided for
informational purposes only by FEI. Copyright 1994-2016 by
FEI, SAS Merignac, France.

*/

#if defined(HX_INTERNAL_CLASS_WARN)
#undef HX_INTERNAL_CLASS_WARN
#define  HX_INTERNAL_CLASS_WARN_HXBASE
#endif
#pragma warning(push)
#pragma warning(disable:4996)

/// @addtogroup hxcore hxcore
/// @{
#ifndef HX_BASE_H
#define HX_BASE_H

#include <hxcore/api.h>

#ifdef _WIN32
// warning C4355: 'this' : used in base member initializer list
#pragma warning( disable: 4355 )
#endif

#include <stdarg.h>
#include <xpand/hxdefs/HxInternalRules.h>
#include <mclib/McDArray.h>
#include <mclib/McInterfaceOwner.h>
#include <mclib/McInterface.h>
#include <amiramesh/HxParseUtils.h>

#include <set>
#include <map>

class HxDisplay;
class SoNode;
class SoSeparator;
class SoTransform;
class SoRenderAreaSwitch;
class SoAction;
class SoFieldSensor;
class SoSensor;
class HxPort;
class HxPlanarModBase;
class HxObject;
class HxData;
class HxSpatialData;
class HxVolumeDataObject;
class HxBasePool;
class HxProjection;
class HxRenderArea;
class SoSwitch;
class HxBasePrivate;
class QString;
class QxBaseNotifier;
class HxPortTypeInfo;

// Units management

template<class T> class QxGlobalDisplayUnitsChangesListener;

struct HxBaseTclMembersPriv;

/**
 * @brief Base class of all data objects, modules, and editors.
 *
 * It provides the following features:
 *
 * @li Reference counting by means of ref() and unref()
 * methods. Instead of manually calling these methods (which can
 * easily lead to errors) you may use smart pointers as provided
 * by the utility class McHandle.
 *
 * @li Run-time type checking as in Open Inventor. These feature are
 * inherited from the base class McTypedObject.
 *
 * @li Tcl-command interface via the virtual parse method. As
 * soon as a @c HxBase object is created a global Tcl command is
 * registered with the same name as the object itself. By invoking
 * this command the parse method is called. This method
 * checks the arguments of the command and triggers the required
 * actions. It is ensured that every base object has a unique name.
 *
 * @li Display of Open Inventor geometry. Derived classes may call the
 * methods showGeom() and hideGeom() in order to show or
 * hide any kind of geometry that they have generated.
 *
 * @li Clipping. Every base objects has its own list of clippers. A
 * clipper is an object of type HxPlanarMod. It can be used
 * to cut any kind of geometry along a 2D clipping plane.
 *
 * @li Coordinate transformations. Any geometry may be displayed in
 * one of three different coordinate systems. These are <em> world
 * coordinates</em>, <em> table coordinates</em>, or <em> local
 * coordinates</em>.
 * The <em> world coordinate system</em> is the system where the
 * camera of the 3D viewer is defined in.
 * The <em> table coordinate system</em> is usually the same as the
 * world coordinate system. However, it might be different if
 * special modules displaying for example the geometry of a
 * radiotherapy device are used. These modules should call the
 * method useWorldCoords() with a non-zero argument in their
 * constructor. Later on they may then call the method
 * HxController::setWorldToTableTransform of the global object
 * @c #theController. In this way they can cause all other objects
 * to be transformed simultaneously.
 * Finally, the <em> local coordinate</em> system is defined by a
 * transformation matrix which can be stored for each object of
 * type HxSpatialData. This matrix can be modified
 * interactively using the transformation editor. The same matrix
 * can be shared between multiple data objects using the
 * method setControllingData(). For example, typically all
 * display modules attached to a data object will share its
 * transformation matrix, so that the geometry generated by these
 * modules is transformed when you transform the data itself.
 *
 * <b> NOTE: </b> <em> If you derive your class from</em> @c HxBase
 * <em> you must use special macros in your class declaration and
 * definition.
 */

class HXCORE_API HxBase : public McInterface {

    MC_ABSTRACT_HEADER(HxBase);

  public:

    /**
     * @brief Flags.
     */
    enum Flags{
        CAN_BE_CLIPPED = 0,             /** Object can be clipped by other modules. Default is true*/
        CAN_COPY_VIEWER_VISIBILITY,     /** Object visibility can be copied from one viewer to another. Checked in HxBase::copyViewerVisibility. Default is true.*/
        CAN_HAVE_CONSOLE_COMPLETION,    /** Object can have console completion. Default is true.*/
        CONSOLE_COMPLETION_ACTIVATED,   /** The console completion is currently activated on object. Default is false.
                                            The console completion will be automatically activated/deactivated when adding/removing an object to/from @c #theObjectPool.*/
        NUM_HXBASE_FLAGS
    };
    
    /**
     * @brief Set a unique label or name for the object.
     *
     * The label is displayed at the object's icon.
     * For every label there is also a Tcl command,
     * which corresponds to the object. The @c setLabel() method might
     * append some digits to the label in order to make it unique.
     * @param label The label name.
     */
    virtual void setLabel(const QString& label);
    
    /** Set a display label for the object. This label does not have to be unique.
        It will be the label displayed in the GUI. */
    virtual void setDisplayLabel( const QString& displayLabel );

    /** Returns the object's display label. */
    inline const QString& getDisplayLabel() const { return theDisplayLabel; }
    

    /**
     * @brief Returns the object's label.
     *
     * @param withDoubleQuotes If it is set to true, a quotation mark is added
     * as first and last character of the label. The default value is false.
     */
    const QString& getLabel(bool withDoubleQuotes=false) const;
    
    /**
     * @brief This is identical to the object's label.
     *
     * @param withDoubleQuotes If it is set to true, a quotation mark is added
     * as first and last character of the label. The default value is false.
     */
    const char* getName(bool withDoubleQuotes=false) const;

    /**
     * @brief Converts a nil terminated C-style string to Camel Case wrt Avizo convention.
     *
     * This class method converts the input parameter to Camel Case with respect to internal conventions.
     * First letter of the result string is lower case.
     * All blank and '-' characters are removed from the string; words those characters were separating
     * are lower cased, excepted their first letter which is upper cased.
     * Other special characters are replaced by '_'; several '_' following one after the other are reduced to only
     * one '_'.
     * First letter following a '_' is upper cased.
     * For example:
     * input string: "Ortho Slice"   *** result: "orthoSlice"
     * input string: "Ortho-Slice"   *** result: "orthoSlice"
     * input string: "OrthoSlice"    *** result: "orthoSlice"
     * input string: "Ortho@Slice"   *** result: "ortho_Slice"
     * input string: "Ortho@@@Slice" *** result: "ortho_Slice"
     *
     * @param name C-style nil terminated string to convert.
     * @return Camel Case version of the input parameter.
     */
    static QString convertToCamelCase( const char* name );

    /**
     * @brief Return the Camel Case label of the object.
     *
     * @return The Camel Case label of the object.
     */
    const QString& getCamelCaseLabel() const;

    /**
     * @brief Indicates whether the object has a Camel Case label or not.
     *
     * This method always returns false in this implementation; it must be overloaded by classes supporting
     * Camel Case labels.
     *
     * @return True only if the Camel Case label is supported by the object, false otherwise.
     */
    virtual bool isCamelCaseSupported() const;

    /**
     * @brief Composes a label from @c otherLabel by replacing the suffix.
     *
     * Example: 
     * - @c otherLabel: motor.am
     * - @c replacement: "-simplified"
     * => "motor-simplified"
     * @param otherLabel The other label.
     * @param replacement The replacement.
     */
    void composeLabel(const QString& otherLabel, const QString& replacement);

    /**
     * @brief Defines in which viewer geometry is visible.
     *
     * The viewer mask
     * of Amira objects can be modified interactively using the orange
     * viewer toggles displayed in the work area. Each bit of the mask
     * is associated with a particular viewer. Note, that even if the
     * viewer mask is set an object may be invisible because the master
     * viewer mask is not set. The master viewer mask is changed whenever
     * the viewer mask of an upstream data object is modified.
     * @param mask The bit mask.
     * @param masterMask The master bit mask (default value is 0xffffffff).
     */
    virtual void setViewerMask(int mask, int masterMask=0xffffffff);

    /**
     * @brief Returns viewer mask ANDed with viewer master mask.
     *
     * This combination decides if an object is actually visible or not.
     * @return The viewer mask anded with viewer master mask.
     */
    int getViewerMask() const;

    /**
     * @brief Returns the viewer mask of the object.
     *
     * @return The viewer mask of the object.
     */
    int getViewerMaskAsIs() const;

    /**
     * @brief Returns the viewer master mask of the object.
     *
     * @return The viewer master mask of the object.
     */
    int getViewerMasterMask() const;

    /**
     * @brief Sets the viewer visibility flag of the viewer @c fromViewerId to @c toViewerId.
     *
     * It enables to show/hide object of the pool for a particular viewer.
     * @param fromViewerId The src viewer id.
     * @param toViewerId The dest viewer id.
     */
    void copyViewerVisibility(unsigned int fromViewerId, unsigned int toViewerId);

    /**
     * @brief Returns the viewer mask, which had been used to initialize \c viewerMask.
     *
     * Note that the default viewer masks differ for different products e.g. Amira, Avizo
     * @return the default viewer mask.
     */
    static int getDefaultViewerMask();

    /**
     * @brief Adds geometry to the viewers.
     * @c node may point to an arbitrary Open
     * Inventor scene graph. This graph will become visible in all
     * viewers defined in the viewer mask. The reference count of @c node
     * will be increased by one. If @c controllData is non-zero the
     * geometry will be transformed by the local transformation matrix of
     * @c controllData.
     * @param node The node.
     * @param controllData The spatial data.
     */
    void showGeom(SoNode *node, HxSpatialData* controllData);

    /**
     * @brief Calls showGeom(node,0); overridden in HxModule, HxEditor.
     *
     * @param node The node.
     */
    virtual void showGeom(SoNode *node);

    /**
     * @brief Removes geometry node from the viewers.
     *
     * The reference count of @c node is decreased by one.
     * @param node The node.
     */
    void hideGeom(SoNode *node);

    /**
     * @brief Clip geometry by planar module.
     *
     *  You can optionally specify in which viewer (@c viewerMask) this object
     *  must be clipped by @c planarMod.
     * @param planarMod The planar mod.
     * @param viewerMask The viewer mask (default value is 0xffff).
     */
    virtual void clipGeom(HxPlanarModBase* planarMod, int viewerMask = 0xffff);

    /**
     * @brief Unset geometry clipping.
     *
     * @param planarMod The planar mod.
     */
    virtual void unclipGeom(HxPlanarModBase* planarMod);

    /**
     * @brief Returns true if an HxPlanarMod is clipping this object.
     *
     * @return true if an HxPlanarMod is clipping this object.
     */
    bool isClipped() const;

    /**
     * @brief Returns viewer mask in which @c planarMod clips this object.
     *
     * @see setViewerMask() for the definition of the viewer mask. 
     * @param planarMod The planar mod.
     * @return 0 in case this object isn't clipped by @c planarMod.
     */
    int getClipperMask(HxPlanarModBase* planarMod);
    
    /**
     * @brief Tcl command interface of the object.
     *
     * @param interpreter The interpreter.
     * @param argc The number of arguments.
     * @param argv The arguments.
     * @return 1 if parse is done without error.
     */
    virtual int parse(Tcl_Interp* interpreter, int argc, char **argv);

    /**
     * @brief Returns the total number of class instances.
     *
     * @return The total number of class instances.
     */
    static int getNumInstances();

    /**
     * @brief Specifies the data object from which the local transformation is taken from.
     *
     * @param data The spatial data.
     */
    void setControllingData(HxSpatialData* data);
    
    /**
     * @brief Returns data object which defines the local coord transform.
     *
     * @return The data.
     */
    HxSpatialData* getControllingData() const;

    /**
     * @brief Hides or shows pin buttons of all ports.
     *
     * @param showIt The show state.
     */
    void showPinButtons( bool showIt );

    /**
     * @brief Returns number of ports.
     *
     * @return The number of ports.
     */
    int getNumPorts() const;

    /**
     * @brief Returns specified port.
     *
     * @param i The port index number.
     * @return The specified port.
     */
    HxPort* getPort(int i) const;
    
    /**
     * @brief Returns port corresponding to given name.
     *
     * @param name The port name.
     * @return The specified port.
     */
    HxPort* getPort( const QString& name) const;

    /**
     * @brief Returns port with the given label.
     *
     * @param label The port label.
     * @return The specified port.
     */
    HxPort* getPortByLabel( const QString& label) const;

    /**
     * @brief Returns the first port corresponding to given type.
     *
     * @param type The port type.
     * @param ignoreInvisible Specifies if invisible ports should be ignored during the search (default value is true).
     * @return The first port corresponding to given type.
     */
    HxPort* getPortOfType( const McTypeInfo *type, bool ignoreInvisible=true ) const;

    // Propagated interface of HxBasePool

    /**
     * @brief Adds a forbidden label, stores a copy of the string.
     *
     * @param forbiddenLabel The forbidden label.
     */
    static void addForbiddenLabel(const char* forbiddenLabel);

    /**
     * @brief Removes a forbidden label.
     *
     * @param forbiddenLabel The forbidden label.
     */
    static void removeForbiddenLabel(const char* forbiddenLabel);

    /**
     * @brief Removes all forbidden labels.
     */
    static void removeAllForbiddenLabels();
    
    /**
     * @brief Needed internally for save network.
     *
     * @param fp The file.
     */
    HxINTERNAL void saveTclMemberVars(FILE* fp);
    
    /**
     * @brief Gets value of tcl member variable with name
     *
     * @param name The name.
     * @param result The associated value.
     * @return True in case of success  and false if the variable did not exist.
     */
    bool getTclMemberVar(const char *name, QString& result) const;
   
    /**
     * @brief sets value of tcl member variable with name
     *
     * @param name The name.
     * @param value The value.
     */
    void setTclMemberVar(const char *name, const char* value);

    /**
     * @brief Save this tcl member variable in network file (standard) or not.
     *
     * @param name The name.
     * @param serialize The serialize value.
     * @return 1 in case of success  and 0 if the variable did not exist.
     */
    HxINTERNAL int setTclMemberVarSerialize(const char *name, bool serialize);


    HxINTERNAL int unsetTclMemberVar(const char *name);
    
    HxINTERNAL int hasTclMemberFunction(const char *name);
    
    HxINTERNAL void undefineTclMemberFunctions();
    
    HxINTERNAL const char* defineTclMemberFunction(const char *name, const char* args,
				const char* body);
    HxINTERNAL void clearTclMembers();
    HxINTERNAL int callTclMemberFunction(const char *name, const char* args, QString& result);

    /**
     * @brief Returns the separator containing all the geometry of the object.
     *
     * If the object does not display any geometry, i.e., if it has never called
     * showGeom(), a null pointer is returned.
     * @return The separator containing all the geometry of the object.
     */
    SoSeparator* getSeparator() const;
    
    /**
     * @brief Checks if VR mode or TeamWork mode is activated.
     */
    virtual void checkRemoteMode();

    /**
     * @brief Get geometry weight.
     */
    HxINTERNAL virtual int getGeometryWeight();
    
    /**
     * @brief Sets or unsets interest in a global polygon offset.
     *
     * Some modules like HxSurfaceIsolines require surfaces to be rendered using a
     * polygon offset towards the back so that lines can be drawn directly
     * onto the surface without z-buffer artifacts. However, since a polygon
     * offset might introduce other rendering artifacts at sharp polygon
     * boundaries it should only be enabled if necessary. Thus, if some
     * module requires polygon offset to be enabled it may request so by
     * calling this method with @c value==true. If polygon offset is no longer
     * needed the method should be called with @c value==false. The method
     * will be called automatically with @c value==false in the destructor
     * of a HxBase object. The method should only be called from the main
     * thread.
     * @param value The polygon offset state.
     */
    void setRequiresPolygonOffset(bool value);

    /**
     * @brief Checks if this object requires a global polygon offset or not.
     *
     * @return True if setRequiresPolygonOffset(true) was called.
     */
    bool requiresPolygonOffset() const;

    /**
     * @brief Returns true if a global polygon offset is needed by at least one
     * module.
     *
     * At least one module must have called setRequiresPolygonOffset()
     * with value==true, otherwise the method returns false.
     * @return True if a global polygon offset is needed by at least one
     * module, false otherwise.
     */
    static bool polygonOffsetRequired();

    /**
     * @brief Restricts the rendering of this module to a specific list of VR screens or viewers.
     *
     * By default, all viewers are allowed to display the representation of this module.
     * @param visibleAreas The visible areas.
     */
    virtual void setDisplayViewers(const std::set<int>& visibleAreas);

    /**
     * @brief Provided for convenience.
     *
     * See @c setDisplayViewers(const std::set<int>& visibleAreas) for more info.
     * @param visibleAreas The visible areas.
     */
    void setDisplayViewers(const McDArray<int>& visibleAreas);

    /**
     * @brief Return the list of of VR screens or viewers which are allowed to display the representation of this module.
     *
     * If the size of the returned list is 0, it means that all viewers are allowed to
     * display the representation.
     * @param visibleAreas The visible areas.
     */
    void getDisplayViewers(std::set<int>& visibleAreas) const;

    /**
     * @brief Sets class managing projection and shadowing options.
     *
     * Warning: @c display will be deleted by this class.
     * @param display The display.
     */
    void setDisplay(HxDisplay *display);

    /**
     * @brief Gets class managing projection and shadowing options.
     *
     * @return The display.
     */
    HxDisplay *getDisplay();

    /**
     * @brief To get the associated Qt notifier which enables to track all Qt signal emitted by HxBase objects.
     *
     * Must be inherited to specify the return type.
     * @return The Qt notifier.
     */
    virtual const QxBaseNotifier* getNotifier() const;

    //@{
    /**
     * @brief Called to notify when a port has been added/removed to/from the HxBase instance.
     *
     * @param port The port.
     */
    void notifyPortAdded(HxPort* port);
    void notifyPortRemoved(HxPort* port);
    //@}


    /**********************/
    /** Units management **/
    /**********************/

    /**
     * @brief Return display units changes listener.
     *
     * @return The display units changes listener.
     */
    QxGlobalDisplayUnitsChangesListener<HxBase>* getDisplayUnitsChangesListener() const;

    /**
     * @brief Called by m_displayUnitsChangesListener when a global display unit has changed.
     *
     * Must be overloaded by derivated classes which need to track display unit changes.
     * @param oldUnit The old unit.
     * @param newUnit The new unit.
     */
    virtual void globalUnitChanged(const int& oldUnit, const int& newUnit);

    /**
     * @brief Called by m_displayUnitsChangesListener when the global coordinates display unit has changed.
     *
     * Must be overloaded by derivated classes which need to track coordinates display unit changes.
     * @param oldUnit The old unit.
     * @param newUnit The new unit.
     */
    virtual void globalCoordinatesUnitChanged(const int& oldUnit, const int& newUnit);

    /**
     * @brief Returns flag value.
     *
     * @param flag The flag.
     * @return The flag value.
     */
    bool getFlag( unsigned int flag ) const;

    /**
     * @brief Sets flag with corresponding value.
     *
     * @param flag The flag.
     * @param value The flag value.
     */
    void setFlag( unsigned int flag, bool value );

    /**
     * @brief Returns the translation text for sourceText, by querying the installed translation files.
     *
     * @param context The context.
     * @param sourceText The source text.
     * @param disambiguation The disambiguation (default value is 0).
     * @return The translation text for sourceText.
     */
    static QString translate( const char * context, const char * sourceText, const char * disambiguation = 0);

    /*
     *  @name On-the-fly port factory
     *
     *  The 4 following function add the notion of on-the-fly port:
     *
     *  Usually, ports are defined as inner object of a class inheriting from HxBase: example:
     *
     *  class HxSomeModule
     *  {
     *  public:
     *      HxPortButtonMenu portButtonMenu;
     *  };
     *
     *  The compiler is aware at compile-time which port belongs to each module and they are
     *  statically allocated. (as inner-objects)
     *
     *  An on-the-fly port is a port that can be added at runtime. Such a port is still listed
     *  in the usual member variables allPorts and will be saved as an usual port. But as they are
     *  instantiated dynamically, the inner variable m_allOnTheFlyPorts keep a list of all on-the-fly
     *  ports that must be deleted by HxBase. Consequently, an on-the-fly port has a life-time bound
     *  to the lifetime of the HxBase used to instantiate.
     *
     */
 
    /**
     * @brief Create a port of determined type.
     *
     * @param className string containing the C++ class name of port to instantiate.
     * @param name Name of the port. (tcl name)
     * @param label of the port. (text displayed by the port Gui)
     *
     * @note This function will throw an exception in 2 cases:
     *           - className is not a recognizable C++ class.
     *           - This class can't instantiate port. (only HxObject and HxEditor could)
     * @return The created port.
     */
    HxINTERNAL HxPort* createOnTheFlyPortFromClassName(const QString& className, const QString& name, const QString& label);

    /**
     * @brief Create a port of determined type.
     *
     * This method is intended to be overridden (HxObject and HxEditor do).
     * By default, it will throw an exception as HxBase can't instantiate a port.
     * @param typeInfo HxPortTypeInfo structure referring to the C++ class name of port to instantiate.
     * @param name Name of the port. (tcl name)
     * @param label of the port. (text displayed py the port Gui)
     *
     * @note This function will throw an exception if this class can't
     *       instantiate port. (only HxObject and HxEditor could)
     * @return The created port.
     */
    HxINTERNAL virtual HxPort* createOnTheFlyPort(const HxPortTypeInfo* typeInfo, const QString& name, const QString& label);

    /**
     * @brief Remove an on-the-fly port.
     *
     * @param The port.
     */
    HxINTERNAL void removeOnTheFlyPort(HxPort* port);

    /**
     * @brief Remove all on-the-fly ports.
     */
    HxINTERNAL void removeAllOnTheFlyPorts();

protected:
    /**
     * @brief Constructor.
     *
     * Constructor is protected - this is an abstract class.
     *
     * To add signals/slots for inherited classes, a new class QxNewClassNotifier
     * which inherits from QxBaseNotifier must be implemented.
     * Then the notifier must be given to the constructor of HxBase, and
     * the function getNotifier() must be overridden.
     * In inherited class, getNotifier() must return the type of the new notifier.
     */
    HxINTERNAL HxBase(McInterfaceOwner* owner = 0, QxBaseNotifier* notifier = 0);

    /**
     * @brief Destructor.
     *
     * Destructor, declared virtual.
     */
    HxINTERNAL virtual ~HxBase();

    /**
     * @brief Derived classes should call this if they want to use world coords.
     *
     * @param trueOrFalse The value.
     */
    void useWorldCoords(int trueOrFalse=1);

    /**
     * @brief Inserts transform nodes into root as necessary.
     */
    virtual void updateTransform();

    // Pointer to the transformation node of the controlling data object.
    SoTransform* globalToLocalTransform;

    // This is zero if the object uses world coords.
    SoTransform* globalToWorldTransform;
    
    // Unique label of the object, also used as Tcl command name.
    QString theLabel;
    QString theLabelWithDoubleQuotes;
    QString theLabelCamelCase;

    // Label that is displayed in the GUI icon, does not need to be unique
    QString theDisplayLabel;

    // Transform matrix of @c controllingData will be used for this object.
    HxSpatialData* controllingData;

    // List of all ports of this object.
    McDArray<HxPort*> allPorts;

    /*
     *  List of all ports created on-the-fly.
     *  All on-the-fly port are listed in allPorts too. The purpose of this vector
     *  is to keep a track of all ports that have been instantiated dynamically
     *  in order to destroy them in the HxBase dtor. Usual port don't need to be deleted
     *  as they are inner-object of some class inheriting from HxBase.
     */
    std::vector<HxPort*> m_allOnTheFlyPorts;

    // This node holds all geometry generated by the object.
    SoSeparator* root;

    // This node holds geometry created by inherited classes and
    // added to this node by using @c showGeom method.
    SoRenderAreaSwitch* raSwitch;

    //Be notified on "on" @c raSwitch field modification
    SoFieldSensor* raSwitchOnSensor;

    // This class holds display properties & ports
    HxDisplay *m_display;

    // Defines in which viewer geometry is visible.
    unsigned short viewerMask;
    unsigned short viewerMasterMask;

    // used by Scaleviz depth composing. 
    // Use zero to prevent a geometry to be rendered on one dedicated slave.
    int m_geometryWeight;

    typedef struct
    {
        SoRenderAreaSwitch* clipSwitch;
        SoFieldSensor* clipSwitchOnSensor;
        int viewerMask;
    } ClipperInfo;

    // List of clipping planes applied to the output of this object.
    std::map<HxPlanarModBase*, ClipperInfo> clipperList;

    /**
     * @brief Interface to Tcl scripting language.
     *
     * @param data The data.
     * @param interpreter The interpreter.
     * @param argc The number of arguments.
     * @param argv The arguments.
     * @return 1 if parse is done without error.
     */
    static int staticParse(ClientData data, Tcl_Interp* interpreter, int argc, char** argv);

    /*
     *  The list of areas in which this module (the scene graph contain within this module)
     *  must be visible.
     *  We use a set to be sure that each element is unique and to benefit of the find capability
     *  of the set.
     */
    std::set<int> m_visibleAreas;

    /**
     * @brief This method is called whenever a projection is activated
     */
    virtual void projectionActivated();

    /**
     * @brief This method is called whenever a projection is deactivated
     */
    virtual void projectionDeactivated();

    /**
     * @brief Checks if the scene graph managed by this instance contains a node.
     *
     * @return True if root or raSwitch contain a node, false otherwise.
     */
    bool containsNode();

    /**
     * @brief Remove @c root from @c #theController and delete @c root and @c raSwitch.
     */
    HxINTERNAL void deleteSceneGraph();

    /**
     * @brief Add @c rootSep to a controller.
     *
     * When adding @c rootSep, it's possible to set its
     * default visibility by setting @c rootViewerMask.
     * This method can be re-implemented in case you would like to use your own controller.
     *
     * @param rootSep The root separator.
     * @param rootViewerMask The root viewer mask (default value is 0xffff).
     */
    virtual void addRootToController(SoSeparator* rootSep, int rootViewerMask = 0xffff);

    /**
     * @brief Remove @c rootSep from a controller.
     *
     * This method can be re-implemented in case you would like to use your own controller.
     * @param rootSep The root separator.
     */
    virtual void removeRootFromController(SoSeparator* rootSep);

    /**
     * @brief The render area switch callback.
     *
     * This callback is activated only when the "setScreen" TCL command or
     * the "setDisplayViewers" method has been called.
     */
    HxINTERNAL static int renderAreaSwitchStaticCB(void* localUserData, HxBase* moduleBase, SoSwitch* traversingSwitch, HxRenderArea* renderArea, SoAction* action);

    /**
     * @brief The render area switch callback.
     *
     * This callback is activated only when the "setScreen" TCL command or
     * the "setDisplayViewers" method has been called.
     *
     * @param traversingSwitch The Transversing switch.
     * @param renderArea The render area.
     * @param action The action.
     */
    virtual int renderAreaSwitchCB(SoSwitch* traversingSwitch, HxRenderArea* renderArea, SoAction* action);

    /**
     * @brief This method is called by @c renderAreaSwitchCB(SoSwitch* traversingSwitch, HxRenderArea* renderArea, SoAction* action)
     *  in order to know which child has to be traversed according to @c visibleAreas.
     */
    HxINTERNAL int handleRenderAreaSwitchTraversing(const std::set<int>& visibleAreas, SoSwitch* traversingSwitch, HxRenderArea* renderArea, SoAction* action);

    /**
     * @brief Use to setting up some necessary OIV cache properties when @c raSwitch is started.
     */
    HxINTERNAL static void raSwitchOnSensorCB(void* data, SoSensor* sensor);

    /**
     * @brief This helper is called by @c setDisplayViewers(const std::set<int>& visibleAreas)
     *
     *  It is in charge of:
     *      -updating the @c visibleAreas list in order to add appropriated SO_SWITCH_NONE or SO_SWITCH_ALL,
     *      -ensuring that the @c root and @c raSwitch exist,
     *      -starting or stopping the @c raSwitch filtering process.
     *  Re-implementing this method is useful when you want to mix the user defined @c m_visibleAreas list with your own one.
     */
    HxINTERNAL void handleNewDisplayViewersList(std::set<int>& visibleAreas);

    /*
     * Array of object's flags. Inherited modules can add custom flags.
     */
    McDArray<bool> m_flags;


    // Units management

    // QxGlobalDisplayUnitsChangesListener object which listen to global display units changes signals.
    QxGlobalDisplayUnitsChangesListener<HxBase>* m_displayUnitsChangesListener;

    /**
     * @brief Initializes the scene graph.
     */
    virtual void initScenegraph();

  private:
    /**
     * @brief Tcl command to parse all interfaces of the object.
     *
     * @param t The interpreter.
     * @param argc The number of arguments.
     * @param argv The arguments.
     * @return 1 if parse is done without error.
     */
    int parseAllInterfaces( Tcl_Interp* t, int argc, char** argv );

    HxBaseTclMembersPriv* tclMembers;

    // private implementation of HxBase.
    HxBasePrivate* m_private;

    QByteArray theLabelQBA;
    QByteArray theLabelWithDoubleQuotesQBA;
    QByteArray theLabelCamelCaseQBA;
    
    /* Private method to set the label and register it as a Tcl command.
        (W)ithout (D)eleting (O)r (U)nique (C)heck, */
    void setLabelWDOUC(const char* label);

  public:
    /**
     * @brief internal use only
     */
    HxINTERNAL void internalSetLabel(const char* label);

    friend class HxPort;
    friend class HxSpatialData;
    friend class HxBasePool;
    friend class HxProjection;
    friend class HxBasePrivate;
};

/*
 *  The following macros are defined for backward compatibilty
 */

/*
 * @page HxBaseRelatedMacros HxBase - related macros
 * Some macros used with classes derived from HxBase.
 * Since HxBase is derived from McTypedObject you <b> must not</b>
 * forget to use one of the macros defined in
 * @ref McTypedObject RelatedMacros "McTypedObject - related macros" in your class declaration and
 * definition. The macros defined <em> above</em> are exactly the same,
 * only use the <b> HX</b> notation.
 * See @ref McTypedObject RelatedMacros "McTypedObject - related macros" or McTypedObject
 * for more details.
 */

/*
 * @name Macros, used in the class declaration
 */
//@{
/* Same as @c MC_ABSTRACT_HEADER(className).*/
#define HX_ABSTRACT_HEADER(className) \
    MC_ABSTRACT_HEADER(className); \
    static inline QString tr(const char *s, const char *c = 0, const char *dummy=0) { dummy = 0; return HxBase::translate(#className, s, c); }
/* Same as @c MC_HEADER(className).*/
#define HX_HEADER(className) \
    MC_HEADER(className); \
    static inline QString tr(const char *s, const char *c = 0, const char *dummy=0) { dummy = 0; return HxBase::translate(#className, s, c); }
/* Same as @c MC_TEMPLATE_HEADER(className).*/
#define HX_TEMPLATE_HEADER(className) \
    MC_TEMPLATE_HEADER(className); \
    static inline QString tr(const char *s, const char *c = 0, const char *dummy=0) { dummy = 0; return HxBase::translate(#className, s, c); }
//@}

/*
 * @name Macros, used in the class definition
 */
//@{
/* Same as @c MC_ABSTRACT_CLASS(className,parentClass).*/
#define HX_INIT_ABSTRACT_CLASS(className,parentClass) \
    MC_INIT_ABSTRACT_CLASS(className,parentClass);

/* Same as @c MC_ABSTRACT_CLASS2(className,parent1,parent2).*/
#define HX_INIT_ABSTRACT_CLASS2(className,parent1,parent2)\
    MC_INIT_ABSTRACT_CLASS2(className,parent1,parent2);

/* Same as @c MC_INIT_CLASS(className,parentClass).*/
#define HX_INIT_CLASS(className,parentClass)\
    MC_INIT_CLASS(className,parentClass);

/* Same as @c MC_INIT_TEMPLATE_CLASS(className,parentClass).*/
#define HX_INIT_TEMPLATE_CLASS(className,parentName) \
        MC_INIT_TEMPLATE_CLASS(className,parentName);

/* Same as @c MC_INIT_CLASS2(className,parent1,parent2).*/
#define HX_INIT_CLASS2(className,parent1,parent2)\
    MC_INIT_CLASS2(className,parent1,parent2);
//@}

/**
 * @brief Set result name.
 *
 * @param[out] result New object name.
 * @param[in]  originalName Original object name (incl. file path & extension).
 * @param[in]  replacement New suffix to use.
 */
extern HXCORE_API void setResultName(QString& result,
    const QString& originalName, const QString& replacement);
    
#endif

/// @}

#pragma warning(pop)

#if defined(HX_INTERNAL_CLASS_WARN_HXBASE)
#undef HX_INTERNAL_CLASS_WARN_HXBASE
#define  HX_INTERNAL_CLASS_WARN
#endif
