/*
Copyright 1999-2016 FEI, SAS
Copyright 1995-2016, ZIB Berlin
ALL RIGHTS RESERVED.

Avizo is a registered trademark of FEI, SAS
Amira is a registered trademark of ZIB.

Avizo and Amira are being jointly developed by Zuse Institute Berlin and FEI, SAS.
HardCopy, MeshViz, VolumeViz, TerrainViz are marks of FEI, SAS
FEI, SAS is a source licensee of OpenGL, Open Inventor
from Silicon Graphics, Inc. OpenGL and Open Inventor are
registered trademark of Silicon Graphics, Inc. All other products and
company names are trademarks or registered trademarks of
their respective companies.

FEI, SAS accepts no liability to any
party for loss or damage caused by errors or omissions or by
statements of any kind in this publication, whether such errors
result from accident, negligence or any other cause. FEI assumes
no liability for incidental or consequential damages arising
from the use of information in this publication. FEI provides no
warranties regarding the information contained in this
publication, whether expressed, implied, or statutory, including
implied warranties of merchantability or fitness for a
particular purpose.

FEI also does not assume any liability for links from this
publication to publications from other companies, organizations
or individuals. FEI does not imply endorsement of any specific
company, organization or individual through offering links to
this content. All information provided within this publication
is subject to change at any time and is provided for
informational purposes only by FEI. Copyright 1994-2016 by
FEI, SAS Merignac, France.

*/

#if defined(HX_INTERNAL_CLASS_WARN)
#undef HX_INTERNAL_CLASS_WARN
#define  HX_INTERNAL_CLASS_WARN_HXCOMPMODULE
#endif
#pragma warning(push)
#pragma warning(disable:4996)

/// @addtogroup hxcore hxcore
/// @{
#ifndef _HX_COMP_MODULE_
#define _HX_COMP_MODULE_

#include <hxcore/HxBatchModule.h>
#include <hxcore/HxData.h>
#include <mclib/McHandle.h>

class HxMasterConnection;

/**
 * @brief Base class for compute modules.
 *
 *  Base class for compute modules like @c HxResample, @c HxArithmetic, or
 *  @c HxMagnitude. Compute modules take one or multiple input objects
 *  and compute one or multiple output objects of type HxData. The
 *  output objects should be registered by calling setResult().
 *  setResult() connects the master connection of the data object
 *  to the compute module, so that the result is downstream with respect
 *  to the compute module. Multiple output objects can be registered
 *  by using an optional id when calling setResult(). Each id defines
 *  a so-called result slot. With each result slot a name can be
 *  associated by calling setNameOfResultSlot().
 *
 *  Compute modules are colored indian red and are usually placed in the
 *  middle region of the object pool. Since this class is derived from
 *  HxBatchModule it is possible to submit a background thread using
 *  HxBatchModule::runBatch().
 *
 */
class HXCORE_API HxCompModule : public HxBatchModule {

  HX_ABSTRACT_HEADER(HxCompModule);

  public:
    /**
     * @brief Associates a custom name with a result id.
     *
     * The name is used when connecting a data object
     * to a compute module in a script or via the graphical
     * user-interface.
     * @param id The result id.
     * @param name The name.
     */
    void setNameOfResultSlot(int id, const char* name);

    /**
     * @brief Registers a data object as a result object.
     *
     * The method first adds the result to the object pool,
     * if it was not already added before.
     * It then connects the master connection of the data
     *  object to this module, thereby possibly disconnecting it from
     *  some other compute module or editor. It also touches the result
     *  using the specified mask and stores the result's touch time, so
     *  that it can be retrieved later on using @c getTouchTime().
     *  It also stores the module's touch time in the result slot.
     *  This is useful in order to decide whether a result can be
     *  recomputed later on. For details see HxObject::canCreateData.
     *  The result is associated with the specified @c id. If a new id
     *  is used a new result slot will be created on the fly.
     * @param id The result id.
     * @param result The result.
     * @param mask The mask (0xffffffff by default).
     */
    void setResult(int id, HxData* result, unsigned int mask=0xffffffff);

    /**
     * @brief Inserts a data object as a result object at the given result-slot
     * index @c pos.
     *
     * All results-slots after the newly created one will be
     * right-shifted.
     * @param pos The index.
     * @param result The result.
     * @param mask The mask (0xffffffff by default).
     */
    void insertResult(int pos, HxData* result, unsigned int mask=0xffffffff);

    /**
     * @brief Remove the result slots at position @c pos.
     *
     * All results slots at the right of the deleted one will be left-shifted.
     * @param pos The index.
     */
    void removeResult(int pos);

    /**
     * @brief Registers a data object as the default result.
     *
     * The default result slot has id==0 and is created
     * for every compute module.
     * @param result The result.
     * @param mask The mask (0xffffffff by default).
     */
    void setResult(HxData* result, unsigned int mask=0xffffffff);

    /**
     * @brief Registers a data object as a result object.
     *
     * A result slot with the given name @c name is used.
     * The parameter @c name must not be 0.
     * @param name The name.
     * @param result The result.
     * @param mask The mask (0xffffffff by default).
     */
    void setResult(const char* name, HxData* result, unsigned int mask=0xffffffff);

    /**
     * @brief Returns the result object associated with the specified id.
     *
     * If no result object is connected to the specified result slot the
     * method returns null.
     * @param id The result id (0 by default).
     */
    HxData* getResult(int id=0) const;

    /**
     * @brief Returns the result object associated with the specified name.
     *
     * If no result object is connected to the specified result slot the
     * method returns 0
     * @param name The name.
     * @return The result object associated with the specified name.
     */
    HxData* getResult(const char* name) const;

    /**
     * @brief Returns the touch time of a result slot.
     *
     * The touch time is set
     * when registering a result using setResult(). Usually the
     * touch time will only be evaluated in an overloaded implementation
     * of HxObject::canCreateData. @see HxData::getTouchTime
     * @param id The result id.
     * @return The touch time of a result slot.
     */
    int getTouchTime(int id=0) const;

    /**
     * @brief Checks if a data object can be used as a result.
     *
     * The method is called before a new result is actually connected
     * to a compute module.
     * For backward compatibility the default implementation
     *  simply returns 1, which means that any data object can be
     *  connected to the module. Derived classes may change this
     *  behaviour by overloading the method. If the result can not
     *  be used as a result for the slot specified by id but for some
     *  other slot, the check method may change the value of id and
     *  still return 1.
     * @param id The result id.
     * @param result The result.
     * @return The checked result.
     */
    virtual int checkResult(int& id, HxData* result);

    /**
     * @brief This method is called after a new result is connected to the
     * module.
     *
     * The id indicates which result slot has been modified.
     * @param id The result id.
     */
    virtual void resultChanged(int id);

    /**
     * @brief Custom fire method.
     *
     * The method simply calls the fire method of the
     *  parent class HxBatchModule. However, before it resets the
     *  variable @c resultTouchTime if at least one port is new. If the
     *  network is saved after a port has changed but before the results
     *  have been recomputed, an overloaded HxObject::canCreateData
     *  method should return false. Note, that the touch times of the
     *  individual result slots as returned by getTouchTime will
     *  not be reset by the fire method.
     * @param shouldCompute If true, the compute method is called.
     */
    virtual void fire(int shouldCompute=1);


    /**
     * @brief Compute modules are #F97C7A.
     *
     * @return The color.
     */
    virtual McColor getIconColor();

    /**
     * @brief Tcl-command interface.
     *
     * @param interpreter The interpreter.
     * @param argc The number of arguments.
     * @param argv The arguments.
     * @return 1 if parse is done without error.
     */
    virtual int parse(Tcl_Interp* interpreter, int argc, char **argv);

    /**
     * @brief Overrides HxObjects's saveUpdate.
     *
     * Saves state of the compute module.
     * @param fp The file.
     */
    virtual void saveUpdate(FILE* fp);

    /**
     * @brief Gets the number of result slots.
     *
     * @return The number of result slots.
     */
    int getNumResultSlots() const;

    /**
     * @brief Returns a free and valid result slot id.
     *
     * @return The number of result slots id.
     */
    int getNewResultSlotID() const;

    /**
     * @brief Sets module to apply transformation of source data to result data.
     *
     * @param apply The apply state.
     */
    void setApplyTransfOfSourceToResult(bool apply);

    /**
     * @brief Gets if transformation of source data should be applied to result data of this module.
     *
     * @return The apply transformation of source to result state.
     */
    bool getApplyTransfOfSourceToResult();

    /**
     * @brief Apply transformation of the source object to the object created by this module.
     */
    void applyTransformToResult();

    /**
     * @brief Updates user interface of the module.
     */
    virtual void update();

    /**
     * @brief Checks if given data object can be reconstructed.
     *
     * @param data The data object.
     * @param createCmd The command.
     * @return True if the given data object can be reconstructed, otherwise false.
     */
      virtual bool canCreateData(HxData* data, QString &createCmd);

  protected:
    /**
     * @brief Constructor.
     *
     * Module may be connected to objects of given type.
     */
    HxCompModule(const McTypeInfo* dataType, McInterfaceOwner* owner = 0, QxObjectNotifier* notifier = 0);

    /**
     * @brief Destructor.
     */
    virtual ~HxCompModule();

    /**
     * @brief Create result associated with the specified name.
     *
     * By default, this method tries to find a port action
     * to trigger a compute.
     * @param name The name.
     * @return The associated result.
     */
    virtual HxData* createResult(const QString & name);

    /* Global touch time of the compute module. This time will be set
        whenever setResult() is called. It will be reset back to -1
        if any port has changed and fire is called. */
    HxINTERNAL int resultTouchTime;
    HxINTERNAL int blockNotify;

    class ResultSlot;
    HxINTERNAL McDArray<ResultSlot*> resultSlots;

    /**
     * @brief Finds the result slot by id.
     *
     * @param id The id.
     * @return The result slot.
     */
    HxINTERNAL ResultSlot* findResultSlot(int id) const;

     /**
     * @brief Finds the result slot from string.
     *
     * @param str The string.
     * @return The result slot.
     */
    HxINTERNAL ResultSlot* findResultSlotFromString(const char* str) const;

     /**
     * @brief Finds the result slot from data.
     *
     * @param data The data.
     * @return The result slot.
     */
    HxINTERNAL ResultSlot* findResultSlotFromData(HxData * data) const;

    HxINTERNAL bool m_applyTransfOfSourceToResult;

    /**
     * @brief Record the result in a HistoryLog bundle.
     *
     * The bundle added will define a new step in the scenario, corresponding
     * to the use of the current module.
     * @param id The number of the result slot.
     * @param res The result data.
     */
    HxINTERNAL void recordNewResultInHistoryLog( int id, HxData* res );

    /**
     * In case of modules that modify the input, this method can be used to record
     * the input modification step in the input HistoryLog.
     * @param inputPort is the connection port to which the input data is connected.
     */
    HxINTERNAL void recordInputUpdateInHistoryLog( HxConnection* inputPort );

    /**
     * Add or update the Apply commands of the resultData's last step in HistoryLog.
     * This should be called after recordNewResultInHistoryLog() or recordInputUpdateInHistoryLog().
     * @param resultData is the result we want to update the HistoryLog.
     * @param customApplyCommand is the custom apply command.
     */
    HxINTERNAL void customizeApplyCommandInHistoryLog( HxData* resultData, QString customApplyCommand );

    /**
    * Updates the State command of the resultData's last step in HistoryLog.
    * This should be called after recordNewResultInHistoryLog() or recordInputUpdateInHistoryLog().
    * @param inputPort is the connection port to which the input data is connected.
    */
    HxINTERNAL void updateInputStateCommandInHistoryLog( HxConnection* inputPort );

    /**
     * @brief Return the touch time.
     *
     * @return the touch time.
     */
    HxINTERNAL int getModuleTouchTime();

    friend class HxMasterConnection;

    /**
     * @brief Indicates whether the range calibration of module input should be copied to the output or not.
     *
     * @return True if the range calibration of module input should be copied to the output, otherwise false.
     */
    bool isRangeCalibrationCopyEnabled() const;
    /**
     * @brief Sets a flag indicates whether the range calibration of module input should be copied to the output or not.
     *
     * @param enable The flag value.
     */
    void enableRangeCalibrationCopy( bool enable );

    /**
     * @brief This method copy the range calibration existing in input to output, scaling and offsetting if required.
     *
     * The copy occurs only if the range calibration is enabled.
     * The copy is done by calling HxRangeCalibrationCompliantData::copy().
     *
     * This method should be overridden by derived class that need to set a scale and/or an offset different
     * from default values.
     * @param input The input.
     * @param output The output.
     * @param scale The scaling value (1.0f by default).
     * @param offset The offset value (0.0f by default).
     */
    virtual void copyRangeCalibration( const HxData* input, HxData* output, float scale = 1.0f, float offset = 0.0f ) const;

    /**
     * @brief Indicates whether the material bundles of module input should be copied to the output or not
     * @return true if copy enabled, false otherwise
     */
    bool isMaterialBundlesCopyEnabled() const;

    /**
     * @brief Sets a flag indicates whether the material bundles of module input should be copied to the output or not
     * @param isEnabled to know if copy of material bundle is enabled or not
     */
    void enableMaterialBundlesCopy( bool isEnabled );

    /**
     * @brief Copy material bundles from module input to output
     * @param output output data of module
     */
    void copyMaterialBundles( HxData* output ) const;

    // Flag indicating whether the range calibration of module input should be copied to the output or not
    bool m_isRangeCalibrationCopyEnabled;

    // Flag indicating whether the material bundle of module input should be copied to the output or not
    bool m_isMaterialBundlesCopyEnabled;

    HxINTERNAL int touchTime;
};

#endif

/// @}

#pragma warning(pop)

#if defined(HX_INTERNAL_CLASS_WARN_HXCOMPMODULE)
#undef HX_INTERNAL_CLASS_WARN_HXCOMPMODULE
#define  HX_INTERNAL_CLASS_WARN
#endif
