/*
Copyright 1999-2016 FEI, SAS
Copyright 1995-2016, ZIB Berlin
ALL RIGHTS RESERVED.

Avizo is a registered trademark of FEI, SAS
Amira is a registered trademark of ZIB.

Avizo and Amira are being jointly developed by Zuse Institute Berlin and FEI, SAS.
HardCopy, MeshViz, VolumeViz, TerrainViz are marks of FEI, SAS
FEI, SAS is a source licensee of OpenGL, Open Inventor
from Silicon Graphics, Inc. OpenGL and Open Inventor are
registered trademark of Silicon Graphics, Inc. All other products and
company names are trademarks or registered trademarks of
their respective companies.

FEI, SAS accepts no liability to any
party for loss or damage caused by errors or omissions or by
statements of any kind in this publication, whether such errors
result from accident, negligence or any other cause. FEI assumes
no liability for incidental or consequential damages arising
from the use of information in this publication. FEI provides no
warranties regarding the information contained in this
publication, whether expressed, implied, or statutory, including
implied warranties of merchantability or fitness for a
particular purpose.

FEI also does not assume any liability for links from this
publication to publications from other companies, organizations
or individuals. FEI does not imply endorsement of any specific
company, organization or individual through offering links to
this content. All information provided within this publication
is subject to change at any time and is provided for
informational purposes only by FEI. Copyright 1994-2016 by
FEI, SAS Merignac, France.

*/

#if defined(HX_INTERNAL_CLASS_WARN)
#undef HX_INTERNAL_CLASS_WARN
#define  HX_INTERNAL_CLASS_WARN_HXCONNECTION
#endif
#pragma warning(push)
#pragma warning(disable:4996)

/// @addtogroup hxcore hxcore
/// @{
#ifndef HX_CONNECTION_H
#define HX_CONNECTION_H

#include <mclib/McDArray.h>
#include <mclib/McVec2i.h>
#include <xpand/hxdefs/HxInternalRules.h>
#include <hxcore/HxPort.h>

class HxCompModule;
class HxEditor;
class QPainter;
class QComboBox;
class QToolButton;
class QIcon;
class QxConnectionGUIUpdater;

// Prototype for connection check methods.
typedef int (*SourceCheck)(void* userData, HxObject* src);

/**
 * @brief Port representing inter-object connections.
 *
 * In contrast to other ports
 * connections usually have no gui elements (derived connections like
 * HxPortColormap again do have gui elements). Instead, connections
 * are listed in the popup menu over the tiny little connection area of
 * the object's icon. A connection port can only be connected to objects
 * of a certain type. Allowed types must be indicated in the constructor
 * or must be added to the list of allowed types using addType().
 */
class HXCORE_API HxConnection : public HxPort {

    HX_PORT(HxConnection);

public:
    /**
     * @brief Constructor without source type.
     *
     * A connection port is created, but the list of allowed source
     * types is not yet specified. An object may call addType()
     * to specify what objects are allowed to connect to the port.
     */
    HxConnection(HxObject*, const QString& name, const QString& label);

    /**
     * @brief Constructor with single source type.
     *
     * Creates a connection port which can be connected to all objects
     * of type @c t. The list of allowed source types may be extended by
     * calling addType().
     */
    HxConnection(HxObject*, const QString& name, const QString& label, const McTypeInfo* t);

    /**
     * @brief Constructor without source type.
     *
     * A connection port is created, but the list of allowed source
     * types is not yet specified. An object may call addType()
     * to specify what objects are allowed to connect to the port.
     */
    HxConnection(HxEditor* editor, const QString& name, const QString& label);

    /**
     * @brief Constructor with single source type.
     *
     * Creates a connection port which can be connected to all objects
     * of type @c t. The list of allowed source types may be extended by
     * calling addType().
     */
    HxConnection(HxEditor* editor, const QString& name, const QString& label, const McTypeInfo* t);

    /**
     * @brief Destructor.
     */
    virtual ~HxConnection();

    /**
     * @brief Returns true if @c src is a valid source.
     *
     * An object is considered
     * to be a valid source if it is derived from one of the types
     * contained in the list of allowed types or if it contains an
     * interface (see HxInterface ) matching one of the types
     * contained in that list. Alternatively, a user-defined check
     * function can be defined using setSourceCheck().
     * @param src The source object.
     * @return True if @c src is a valid source, otherwise false.
     */
    int validSource(HxObject* src);

    /**
     * @brief Establishes connection.
     *
     * This method calls HxObject::notifyConnect() on the connected object (i.e. src) if
     * the notification on connection is enabled (cf. enableNotificationOnConnect() / isNotificationOnConnectEnabled()).
     * @param src The connected object.
     * @return True if the connect is enabled, otherwise false.
     */
    virtual bool connect(HxObject* src);

    /**
     * @brief Enables/disables notification on connection. See connect() for details.
     *
     * @param flag Notification on connection state.
     */
    void enableNotificationOnConnect(bool flag);

    /**
     * @brief Returns true if the notification on connection is enabled. See connect() for details.
     *
     * @return True if the notification on connection is enabled, otherwise false.
     */
    bool isNotificationOnConnectEnabled() const;


    /**
     * @brief Removes connection.
     *
     * @param disconnectedDuringConnection Disconnected during connection state (default value is false).
     */
    virtual void disconnect(bool disconnectedDuringConnection=false);

    /**
     * @brief Returns object the port is connected to.
     *
     * @return The object the port is connected to.
     */
    HxObject* getSource() const;

    /**
     * @brief Returns pointer to source object or interface of type @c t.
     *
     * For more information about interfaces refer to HxInterface .
     * @param t The type.
     * @return Null if the port is not connected to any source
     * object or if the source object is not of type @c t and if it not
     * contains an interface of type @c t.
     */
    McTypedObject* getSource(const McTypeInfo* t) const;

    /**
     * @brief Returns true if the connection is optional and false if it's mandatory.
     *
     * Default is false (mandatory).
     * @return True if the connection is optional and false if it's mandatory.
     */
    bool isOptional();

    /**
     * @brief Sets a connection mandatory \a inFlag false or optional \a inFlag
     * true.
     *
     * Default is false (mandatory).
     *
     * During the deletion of a HxObject the corresponding owners of
     * all its downstream connections will be deleted automatically
     * if they are not of type HxData. However if a  downstream connection
     * is optional it will only be disconnected but not deleted.
     *
     * Imagine you have a display module additionally connected to a 'SelectROI' module
     * and you don't want to have the display module deleted whenever the 'SelectROI' module
     * will be deleted. In that particular case you would set the connection of the
     * 'SelectROI' module to optional (true).
     * @param inFlag The optional state.
     */
    void setOptional(bool inFlag);

    /**
     * @brief Returns true if the connection can be edited.
     *
     * Editing means that the
     * user is able to pick the connection line displayed in the object
     * pool and drag it onto some other input object. Non-editable
     * connections are displayed in grey instead blue.
     * @return True if the connection can be edited.
     */
    HxINTERNAL bool isEditable() const;

    /**
     * @brief Allows or disallows interactive editing of a connection.
     *
     * @param onOff The state.
     */
    void allowEditing(bool onOff);

    /**
     * @brief Sets visibility flag.
     *
     * If a connection is invisible (@c cnxVisibility), no connection
     * line will be drawn in the object pool. Sometimes this might be
     * useful for clarity.
     * Use the parameter @c popupCnxItemVisibility to never display
     * the item representing a connection within the popup menu of a module
     * even if a connection has been done.
     * If you want to see the connection item when a module is connected to this connection
     * port and, hide the connection item when a module isn't connected,
     * use @c setVisible instead.
     * On default, all connections will be visible.
     */
    void setVisibility(int cnxVisibility, bool popupCnxItemVisibility=true);

    /**
     * @brief Tell if the item representing a connection within the popup menu of a module
     * even if a connection has been done.
     *
     * Mainly used by @c #theObjectPool.
     */
    HxINTERNAL bool isPopupCnxItemVisible() const;

    /**
     * @brief Sets tightness flag.
     *
     * If a connection is marked as tight, no
     * connection line will be drawn in the object pool. Instead, the
     * icon of the object the connection belongs to will be drawn
     * right under the icon of the source object.
     * @param onOff The state.
     */
    void setTightness(bool onOff);

    /**
     * @brief Returns tightness flag.
     *
     * @return The tightness flag value.
     */
    int isTight() const;

    /**
     * @brief Adds @c t to the list of allowed source types.
     *
     * @param t The type.
     */
    void addType(const McTypeInfo* t);

    /**
     * @brief Clears the allowed source types list.
     */
    void clearTypes();

    /**
     * @brief Sets alternative source check method.
     *
     * @c SourceCheck is defined in
     * the following way:
     * @code
     * typedef int (*SourceCheck)(void* userData, HxObject* src);
     * @endcode
     * The source check method should return 1 if the port can be
     * connected to the object pointed to by @c src. Otherwise, it should
     * return 0.
     */
    void setSourceCheck(SourceCheck cb, void* userData);

    /**
     * @brief Returns current source check callback with its associated userData.
     *
     * @param userData The user data.
     * @return the current source check callback with its associated userData.
     */
    SourceCheck getSourceCheck(void*& userData);

    /**
     * @brief Notifies the connection that its source filter has been changed.
     */
    void sourceCheckChanged();

    /**
     * @brief Tcl command interface.
     *
     * @param interpreter The interpreter.
     * @param argc The number of arguments.
     * @param argv The arguments.
     * @return 1 if parse is done without error.
     */
    virtual int parse(Tcl_Interp* interpreter, int argc, char **argv);

    /**
     * @brief Overloaded save state method.
     *
     * @param fp The file.
     */
    virtual void saveState(FILE* fp);

    /**
     * @brief Returns current state in ascii string.
     *
     * @param state The current state in ascii string.
     */
    virtual void getState(QString& state) const;

    /**
     * @brief Inherited from HxPort.
     * To restore the current state from a string.
     * @internal
     * @param state could contains one of the following tokens:
     *     connect <string-sourceName>
     *     disconnect
     * To set the value in a brief way, the @c setState method can be called directly like this:
     *     setState <string-sourceName>
     * @return 1 if there are no error, otherwise 0.
     */
    virtual int setState(const QString& state);

    /**
     * @brief Draw connection line
     *
     * @param p The painter.
     * @param connectors The connectors.
     */
    HxINTERNAL virtual void drawConnection(QPainter* p, std::map<HxObject*, McDArray<std::pair<QPointF, QColor> > >& connectors);

    /**
     * @brief Check if connection is hit by x, y coordinates.
     *
     * @param x,y The coordinates.
     * @return If the line isn't hit at all 0 is returned. If the anchor point at the source object
     * is hit the method returns -1. Otherwise it returns a number
     * indicating the proximity between the point and the line. The
     * larger the number the closer the point and the line.
     */
    HxINTERNAL virtual int intersect(int x, int y);

    /**
     * @brief Generate graphical feedback when editing connection.
     */
    HxINTERNAL void feedback(QPainter*,int x, int y, HxObject* object);

    /**
     * @brief Called when finishing edit session.
     *
     * @param object The object.
     */
    HxINTERNAL void confirm(HxObject* object);

    /**
     * @brief To know if the connection gui must be displayed before the others ports.
     *
     * @return True if the connection gui must be displayed before the others ports, otherwise false.
     */
    HxINTERNAL int mustBeDisplayedBeforeOtherPorts() const;

    /**
     * @brief To get the sources list combo box.
     *
     * @return The sources list combo box.
     */
    HxINTERNAL QComboBox* getSourcesListCB() const;

    /**
     * @brief To get the "Go to" tool button.
     *
     * @return The "Go to" tool button.
     */
    HxINTERNAL QToolButton* getGoToCurrentSourceBtn() const;

    /**
     * @brief Sets the visibility of the "Go to" tool button.
     *
     * If the "Go to" button don't exist, the visibility is saved and applied when the button is created.
     * @param visible The visible state.
     */
    HxINTERNAL void setVisibleGoToCurrentSourceBtn(bool visible);

    /**
     * @brief Return the visibility of the "Go to" tool button.
     *
     * If the "Go to" button don't exist, it's the saved visibility which is returned.
     * @return The visibility of the "Go to" tool button.
     */
    HxINTERNAL bool isVisibleGoToCurrentSourceBtn();

    /**
     * @brief To get the associated GUI updater.
     *
     * @return The associated GUI updater.
     */
    HxINTERNAL QxConnectionGUIUpdater* getGUIUpdater() const;

    /**
     * @brief Sets the label.
     *
     * @param text The label.
     */
    virtual void setLabel(const QString & text);

    /*
     * Type info array type added for convenience.
     */
    typedef McDArray<const McTypeInfo*> TypeInfoArray;

    /**
     * @brief Return the list of allowed object types to connect to this port.
     *
     * @return The list of allowed object types to connect to this port.
     */
    const TypeInfoArray& getValidTypes() const;

    /**
     *  @brief Copy values from refPort and update GUI.
     *
     *  @param refPort The reference port.
     */
    virtual void initFrom(HxPort* refPort);

    /**
    *  @brief Control if the connection gui must be displayed before the others ports.
    *
    *  @param force If true (default), the connection is displayed first.
    */
    inline void setMustBeDisplayedBeforeOtherPorts(bool force = true) {
        m_mustBeDisplayedBeforeOtherPorts = force;
    }

    HxINTERNAL virtual void createExternal(HxObject *object, int buildWidget);

    HxINTERNAL virtual void removeExternal(HxObject *object);

    /**
     * @brief Sets the flag indicating if this connection accepts data connected to a downstream connection of its owner.
     *
     * @param cyclicConnectionAccepted The cyclic connection accepted state.
     */
    void setAcceptCyclicConnection( const bool cyclicConnectionAccepted );

    /**
     * @brief Indicates if this connection accepts data connected to a downstream connection of its owner.
     */
    bool acceptCyclicConnection() const;

    /**
     * @brief Set the connection as savable (or not).
     *
     * @param canSaveConnection tells if the connection can be saved.
     */
    HxINTERNAL void setSaveConnection( bool canSaveConnection );

    /**
     * @brief Tells if the connection can be saved.
     *
     * @return true if the connection can be saved in any case, false otherwise.
     */
    HxINTERNAL bool canSaveConnection() const;

protected:
    bool editable;
    bool optional;
    bool notificationOnConnectEnabled;
    bool tight;
    HxObject* connected;
    TypeInfoArray typeList;
    void* userData;
    SourceCheck sourceCheck;

    /*
     * To know if the object connection item within the object popup menu in the pool is visible or not.
     * This flag can be set by setVisibility(int cnxVisibility, bool modulePopupCnxItemVisibility).
     */
    bool m_connectionPopupItemVisibility;

    /*
     * To know if the connection line in the pool is visible or not.
     * This flag can be set by setVisibility(int cnxVisibility, bool modulePopupCnxItemVisibility).
     */
    unsigned int m_connectionLineVisibility:1;

    /** 
     * @brief The gui is composed of a combo box (list of available sources) and
     * a "Go to" tool button used to select the current source.
     *
     * @param w The widget.
     */
    virtual void createWidget(QWidget* w);


    HxINTERNAL static int defaultSourceCheck(void* userData, HxObject* source);

    /*
     * To know if the connection gui must be displayed before the others ports.
     * Default value is true. Inherited connections (like HxPortColormap, HxPortTime)
     * which have their own gui must set this flag to false.
     */
    bool m_mustBeDisplayedBeforeOtherPorts;

    /*
     * The sources list combo box.
     */
    QxGuardedPtr<QComboBox> m_sourcesListCB;

    /*
     * The "Go to" tool button.
     */
    QxGuardedPtr<QToolButton> m_goToCurrentSourceBtn;

    /*
     * To know if the goToCurrentSourceBtn must be displayed.
     */
    bool m_showGoToCurrentSourceBtn;

    /*
     * The "Go to" tool button icon.
     */
    static QIcon* s_goToCurrentSourceIcon;

    /** 
     * @brief Common init.
     */
    void commonInit();

    /*
     * Associated object pool listener used to update the gui.
     */
    QxConnectionGUIUpdater* m_guiUpdater;

    /** 
     * @brief Gui callback method.
     */
    virtual void guiCallback();


private:

    /*
     * @brief Add quotes surrounding the given string to protect spaces if necessary.
     *
     * @str the string to protect with quotes.
     * @return the protected string
     */
    static QString addSurroundingQuotes( const QString& str );

    // Flag indicating if this connection accepts data connected to a downstream connection of its owner
    bool m_acceptCyclicConnection;

    /**
     * @brief Tells if the connection can be saved.
     */
    bool m_canSaveConnection;

};

template<class TYPE>
TYPE* hxconnection_cast (const HxConnection& connection) {
    return static_cast<TYPE*> (connection.getSource(TYPE::getClassTypeId()));
}

#endif


/// @}

#pragma warning(pop)

#if defined(HX_INTERNAL_CLASS_WARN_HXCONNECTION)
#undef HX_INTERNAL_CLASS_WARN_HXCONNECTION
#define  HX_INTERNAL_CLASS_WARN
#endif
